%!octave 
%
%
%       OS : GNU/Linux 4.2.5
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: cn_backprop_grad.m,v 1.18 2016/04/21 14:06:58 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% градиент функции ошибки сети
%
% L - учебный набор 
% W - матрица весов
% cfg - описание сети
% dlm - маска dropout для слоёв
%
function retval=cn_backprop_grad(L,W,cfg),

   % фаза прямого распространения
   s = cn_state( L{1}, W, cfg ); % состояния нейронов всех слоёв

   % фаза обратного распространения
   er = cn_errors(s,L{2},W,cfg); % ошибки сети на учебном наборе

   g = cn_grad(s,er,cfg); % градиент ф-ции ошибки сети

   %g = grad_norm(g,cfg); 

   retval = g;

endfunction




% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=cn_errors(l,y,W,cfg),
   ltp=2; % номер ячейки cfg с типом ф-ции активации
   layers_numb = length(cfg) ; % количество слоёв сети, включая вход
 
   er = {} ; % ошибка слоёв
  
   er{layers_numb-1} = cn_loss_drvt( l{end}, y, cfg ) ; % ошибка сети (выходного слоя)
  
   for j=layers_numb:-1:3, % ошибки скрытых слоёв
      er{j-2} = cn_errors_layer(l,er{j-1},W,cfg, j );
   endfor

   retval=er;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=cn_errors_layer(l,er,W,cfg,n),
   ltp=1; % номер ячейки с типом слоя 

   switch cfg{n}{ltp}

	   case "mlp"
	     s = error_mlp( l,er,W,cfg,n );

	   case "conv"
		 s = error_conv( l,er,W,cfg,n );

       case "mpool"
		 s = error_mpool( l,er,W,cfg,n );

	  otherwise
		 printf("[E] cn_errors_layer: неизвестный тип слоя %s \n",cfg{n}{ltp}); 
		 exit -1;

   endswitch

   retval = s ;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=error_mlp(l,er,W,cfg,n),
   lac=3; % номер ячейки cfg с типом активации

   s = ( (W{n-1}{1})' * er ) ; 

   df = cn_activate_drvt( l{n-1}, cfg{n-1}{lac} ) ; % значения производной ф-ции активации 
   if ndims(df) > 2, s = reshape(s,size(df)) ; endif
   
   retval = s  .* df ; 
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = error_mpool(l,er,W,cfg,n),
   lsz = 2 ; % номер ячейки cfg с размером блока выборки (max pooling)
   lac = 3 ; % номер ячейки cfg с типом активации

   sb = cfg{n}{lsz} ; % размер блока выборки (max pooling)
   se = size(er); % размеры входного пакета матриц 
     
   %o = er .* W{n-1}{1} + W{n-1}{2};
   o = er ;

   o = kron(o,ones(sb)); % увеличиваем размер в sb раз, дублируем значения ячеек блоками размера sb

   se(1:2) = se(1:2) .* sb ; % размер увеличеного набора матриц
   o = reshape(o,se) ; % перепаковываем 

   sl=size(l{n-1}); % размер входных карт слоя 
   o=o(1:sl(1),1:sl(2),:,:); % обрезаем лишние элементы под размер карт

   df = cn_activate_drvt( l{n-1}, cfg{n-1}{lac} ) ; % значения производной ф-ции активации 

   retval = o .* df ;

endfunction



% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = error_conv( l,er, W, cfg, n ),
   lac = 3 ; % номер ячейки cfg с типом активации
   lsz = 2 ; % номер ячейки cfg с размером слоя 
   lmp = 4 ; % номер ячейки cfg с матрицей связей слоёв

   link_matr = cfg{n}{lmp} ; % номера выходных карт для обратного распространения

   out_numb_seq = link_numb_seq( link_matr' ) ; % номера выходных карт для обратного распространения

   % map_in_numb = length(out_numb_seq) ; % количество входных карт
    
   szl = size(l{n-1}) ; % размер входа
   map_in_numb = szl(3) ; % количество входных карт
   ex_num = szl(4) ; % количество примеров

   s = zeros(szl) ; % результат "обратных" свёрток

   K = W{n-1}{1} ; % ядра свёрток

   for j=1:map_in_numb, % для всех входных карт
      
      map_out_numb = length(out_numb_seq{j}) ; % количество выходных карт, в которых участвует вход j 

	  if length(szl)>3,
		 sj = zeros( [ szl(1:2), map_out_numb , szl(4) ] ); % "обратные" свёртки для входной карты j
	  else
		 sj = zeros( [ szl(1:2), map_out_numb , 1 ] ); % "обратные" свёртки для входной карты j
	  endif

	  for i=1:map_out_numb, % для выбранных выходных карт 
	     out_num = out_numb_seq{j}(i) ; % номер выходной карты для входа j
         nk = sum( link_matr(1:j,out_num) ) ; % номер ядра свёртки для выхода out_num и входа j
          
         bK = rot180(K{out_num}(:,:,nk)) ; % ядро для "обратной" свёртки
		 %sj(:,:,i,:) = convn( er(:,:,out_num,:),  bK, "full" ) ; % выполнить "обратную" свёртку 

		 for q=1:ex_num,
		 	sj(:,:,i,q) = conv2( er(:,:,out_num,q),  bK, "full" ) ; % выполнить "обратную" свёртку 
		 endfor

	  endfor

	  s(:,:,j,:) = sum(sj,3);  % частичные суммы выходных карт
   endfor

   df = cn_activate_drvt( l{n-1}, cfg{n-1}{lac} ) ; % значения производной ф-ции активации 
   retval = s .* df ;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=cn_grad(l,er,cfg),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации
   g = {} ; % градиент ошибки сети
   
   for j=1:length(er),
	  g{j} = cn_grad_layer(l,er,cfg,j) ;
   endfor

   retval=g;

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = cn_grad_layer(l,er,cfg, n ),
   ltp=1; % номер ячейки с типом слоя 

   switch cfg{n+1}{ltp}

	   case "mlp"
	     s = grad_mlp( l, er, cfg, n );

	  case "conv"
		 s = grad_conv( l, er, cfg, n );
  
	  case "mpool"
		 s = grad_mpool( l, er, cfg, n );

	  otherwise
		 printf("[E] cn_grad_layer: неизвестный тип слоя %s \n",cfg{n+1}{ltp}); 
		 exit -1;

   endswitch

   retval = s;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=grad_mpool(l,er,cfg,n),
   lac = 3 ; % номер ячейки cfg с типом ф-ции активации
  
   map_out_numb = size( er{n}, 3 ) ; % количество выходных карт
   
   w = {} ; w{n} = { ones([1,1,map_out_numb]), zeros([1,1,map_out_numb]) } ;
   o = cn_layer( l{n}, w, cfg, n+1 ) ;
   o = cn_activate(o,cfg{n+1}{lac})  ;

   s = er{n} .* o ;

   g={};

   % g{1} = sum( sum(sum(s,1),2)./ prod(size(s)(1:2)) ,4) ;
   % g{2} = sum(sum(sum(er{n},1),2)./ prod(size(er{n})(1:2)),4) ;

   g{1} = sum( sum(sum(s,1),2),4)(:) ;
   g{2} = sum(sum(sum(er{n},1),2),4)(:) ;

   % g{1} = g{2} = 0.0 ; % FIXME: этот слой выполняет только выборку максимальных элементов по ячейкам

   retval = g ;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=grad_mlp(l,er,cfg,n),
   lac = 3 ; % номер ячейки cfg с типом ф-ции активации
   g = {} ; % градиент ошибки сети
   
   o = cn_activate(l{n},cfg{n}{lac})  ;

   if ndims(o) > 2, o = data_unpack_img(o); endif

   g{1} = er{n} * o' ; 
   g{2} = sum(er{n},2) ;

   retval = g;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = grad_conv( l,er, cfg, n ),
   lmp = 4 ; % номер ячейки cfg с матрицей связей слоёв
   lac=3; % номер ячейки cfg с типом активации
   lsz=2; % номер ячейки cfg с размером слоя

   link_map = link_numb_seq( cfg{n+1}{lmp} ) ; % связи входных и выходных карт
   map_out_numb = length(link_map) ; % количество выходных карт
   ex_numb = size(l{n+1},4) ; % количество учебных примеров

   o = cn_activate(l{n},cfg{n}{lac})  ; % вход слоя

   szk=cfg{n+1}{lsz}; % размеры ядер свёртки

   gk={}; % значение градиента по ядрам свёртки
   gb = zeros([map_out_numb,1]) ; % значение градиента для сдвига 
   for j=1:map_out_numb, % для каждой выходной карты
	  sj = zeros( [ szk , length( link_map{j}) ,ex_numb] );
	  for i=1:ex_numb, % для всех примеров 
		 % выполнить свёртки входных карт с помощью карты ошибки
		 ern = rot180( er{n}(:,:,j,i) ) ;
		 %oi = o(:,:,link_map{j},i) ;
		 %sj(:,:,:,i) = rot180( convn( oi,  ern, "valid" ) ) ;
% FIXME 
		 for q=1:length(link_map{j}),
			oi = o(:,:,link_map{j}(q),i) ;
			sj(:,:,q,i) = rot180( conv2( oi,  ern, "valid" ) ) ;
		 endfor

	  endfor
	  % результат суммировать по примерам
      gk{j} = sum(sj,4); % градиент для ядер свёртки выходной карты j
	  gb(j) = sum(sum(sum(gk{j},1),2),3); % градиент для сдвига 
   endfor

   retval = { gk,gb } ;

endfunction


