%!octave 
%
%      
%      
%       OS : GNU/Linux 4.3.3-3-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: cn_init.m,v 1.13 2016/04/21 14:06:58 mechanoid Exp $

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% инициализация параметров сети
%
% cfg - конфигурация сети 
%
function retval = cn_init( cfg ),
   lsz=2; % номер ячейки с размером слоя 
   
   s = cfg{1}{lsz}; % размер первого (входного) слоя сети

   W={}; % веса сети
   for i=1:length(cfg)-1, % для всех слоёв
	  [ Wi, s ] = init_layer( cfg,i+1, s ) ;
	  W{i} = Wi;
   endfor

   retval = W ;

endfunction 


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% инициализация параметров слоя
%
% cfg - конфигурация сети 
% n - номер слоя
% W - параметры предыдущих слоёв
% s - размер предыдущего слоя
%
function [ W, s ]= init_layer( cfg, n, s ),
   ltp=1; % номер ячейки с типом слоя 

   switch cfg{n}{ltp}

   	  case "conv"
		 [ W,s ] = init_layer_conv( cfg, n, s );

	  case "mpool"
		 [ W,s ] = init_layer_mpool( cfg, n, s );

	  case "mlp"
	     [ W,s ] = init_layer_mlp( cfg, n, s );
	  
	  otherwise
		 printf("[E] init_layer: неизвестный тип слоя %s \n",cfg{n}{ltp}); 
		 exit -1;

   endswitch


endfunction 




% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function [ W, s ] = init_layer_mlp( cfg, n, s ),
   lsz=2; % номер ячейки с размером слоя 

   sz_pl = prod(s); % количество нейронов в пред.слое

   W={ small_random( [ cfg{n}{lsz}, sz_pl ], 1e-3 ) ,  zeros(cfg{n}{lsz},1)  };

   s = cfg{n}{lsz}; % размер слоя

endfunction 

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function [ W, s ] = init_layer_mpool( cfg, n, s ),
   lsz=2; % номер ячейки с размером слоя 

   numb_maps = s(3);
 
   % W = { small_random( [ numb_maps,1 ] ), zeros( [ numb_maps,1 ] )   } ;
   W = { ones( [ numb_maps,1 ] ), zeros( [ numb_maps,1 ] )   } ;
   % W = { 1.0, 0.0 } ; % этот слой выполняет только выборку максимальных элементов по ячейкам

   s(1:2) = ceil(s(1:2)./cfg{n}{lsz}) ;

endfunction 

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% инициализация параметров свёрточного слоя
%
% cfg - конфигурация сети 
% n - номер слоя
% W - параметры предыдущих слоёв
% s - размер предыдущего слоя
%
function [ W, s ] = init_layer_conv( cfg, n, s ),
   lsz = 2 ; % номер ячейки cfg с размером слоя 
   lmp = 4 ; % номер ячейки cfg с матрицей связей слоёв

   link_map = link_numb_seq( cfg{n}{lmp} ) ; % номера входных карт для частичных сумм

   map_out_numb = length(link_map); % количество результирующих карт
   sz_kern = cfg{n}{lsz} ; % размер ядра свёртки
   
   % одна карта формируется свёрткой частичной суммы входных карт и сдвигом: act( conv(psum(m),k_i)+b )

   k={}; % ядра свёрток
   for i=1:map_out_numb, % для всех выходных карт
      kern_numb = length(link_map{i} ) ; % количество ядер свёртки
	  k{i} = small_random( [ sz_kern, kern_numb ], 1e-3 ) ;
   endfor

   W = { k, zeros( [ map_out_numb,1 ] ) } ;

   sz_map = s(1:2) - sz_kern + 1 ; % размер выходной карты

   s = [ sz_map, map_out_numb ] ; % размеры слоя n

endfunction 


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = small_random(s,a),
   %a=1e-2;
   retval = normrnd(zeros(s),ones(s)*a);
endfunction



