%!octave 
%      
%       OS : GNU/Linux 4.2.5
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: cn_layer.m,v 1.16 2016/04/21 14:06:58 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = cn_layer( X, W, cfg, n ),
   ltp=1; % номер ячейки с типом слоя 

   switch cfg{n}{ltp}

	  case "conv"
		 s = layer_run_conv( X, W, cfg, n );

	  case "mlp"
	     if ndims(X) > 2, X = data_unpack_img(X); endif
	     s = layer_run_mlp( X, W, cfg, n );

	  case "mpool"
	     s = layer_run_mpool( X, W, cfg, n );
	  
	  otherwise
		 printf("[E] cn_layer: неизвестный тип слоя %s \n",cfg{n}{ltp}); 
		 exit -1;

   endswitch

   retval = s;

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=layer_run_conv( X, W, cfg, n ),
   % ex_numb = size(X,4); % количество учебных примеров
   lmp = 4 ; % номер ячейки cfg с матрицей связей слоёв

   link_matr = cfg{n}{lmp} ; % матрица связи входных и выходных карт
   link_map = link_numb_seq( link_matr ) ; % индексы связей входных и выходных карт
   map_out_numb = length(link_map); % количество выходных карт

   K = W{n-1}{1} ; % ядра свёртки
   b = W{n-1}{2} ; % сдвиги
   
   sz_kern = size(K{1})(1:2) ; % размер ядра
   sz_map_out = size(X)(1:2) - sz_kern + 1 ; % размер выходной карты

   if ndims(X)<4,
	  ex_numb=1; % вход состоит из одного примера 
   else
	  ex_numb=size(X,4);
   endif

   o = zeros([sz_map_out,map_out_numb,ex_numb]); % результат

   for j=1:map_out_numb, % для всех выходных карт
	  nmaps = link_map{j} ; % номера задействованных входных карт
	  maps_numb = length(nmaps) ; % количество задействованных входных карт
	  oji=zeros([sz_map_out,maps_numb,ex_numb]); % свёртки для выхода j
	  for i=1:maps_numb, % для входных карт nmaps выхода j
		 % выполнить свёртки соответствующими ядрами
		 oji(:,:,i,:) = convn( X(:,:,nmaps(i),:), K{j}(:,:,i), "valid" ) ;

		 % for q=1:ex_numb,
		 %	oji(:,:,i,q) = conv2( X(:,:,nmaps(i),q), K{j}(:,:,i), "valid" ) ;
		 % endfor

	  endfor
	  % вычислить сумму результата и выполнить сдвиг
	  o(:,:,j,:) = sum(oji,3) + b(j);
   endfor

   retval=o;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=layer_run_mlp( X, W, cfg, n ),
   s = ( W{n-1}{1} * X + W{n-1}{2} ) ;
   retval=s;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=layer_run_mpool( X, W, cfg, n ),
   lsz=2; % номер ячейки cfg с размером ячёйки выборки элемента (max pooling)

   s=cfg{n}{lsz}; % размер блока subsumpling

   xe=matr_expand(X,s) ; % дополняем матрицу до нужного размера

   % выполняем выборку максимального элемента в блоках (max pooling)
   xr = max_pooling_rows(xe,s(1)) ; % выборка максимальных элементов по строкам

   xr = permute(xr,[2,1,3,4]) ; % транспонируем пакет матриц
   xr = max_pooling_rows(xr,s(2)) ; % выборка максимальных элементов по столбцам
   xr = permute(xr,[2,1,3,4]) ; % транспонируем пакет матриц
  
   if ndims(xr) < 4,
	  s=size(xr);
	  xr=reshape(xr,[s(1:2),1,s(3)]) ;
   endif

   a=W{n-1}{1};
   b=W{n-1}{2};
   
   %a=permute(a,[2 3 1 4]);
   %b=permute(b,[2 3 1 4]);
   %xr = xr.*a + b ;

   for i=1:size(xr,3),
	  xr(:,:,i,:) = xr(:,:,i,:) * a(i) + b(i);
   endfor


   retval=xr;

endfunction 

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% дополняем матрицу до соответствия размерам блока 
%
% x - входная матрица данных
% s - размер блока
%
function retval=matr_expand(x,s),
   sz_matr = size(x)(1:2); % размер исходной матрицы
   step_numb = floor(sz_matr./s); % количество шагов (блоков)
   sz_tail = sz_matr - step_numb.*s ;  % размер остатка
   sz_exp = (s - sz_tail) .* (sz_tail>0) ; % размер дополнения 

   xr=x;

   if(sz_tail(1)>0), % дополняем снизу, копируем последнюю строку 
      xr = [ xr ; xr(end,:,:,:) .* ones(sz_exp(1),1) ] ; 
   endif

   if(sz_tail(2)>0), % дополняем справа, копируем последний столбец 
	  xr = [ xr, xr(:,end,:,:) .* ones(1,sz_exp(2)) ] ; 
   endif

   retval=xr;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% выбор максимальных элементов в блоках строк размера s
% 
% x - входная матрица данных
% s - количество строк в блоке
%
function retval=max_pooling_rows(x,s),
   step_numb=floor(size(x,1)/s); % количество шагов (блоков)
   j1=0; % последняя строка блока выборки

   xr=[]; % результат
   for i=1:step_numb, % для всех наборов рядом стоящих строк по sz_cell(1) штук
	  j0=j1+1; % первая строка блока выборки
	  j1=j0-1+s; % последняя строка блока выборки

	  switch ndims(x),
		 case 2
			xr(i,:) = max( x( j0:j1 ,:) );
		 case 3
			xr(i,:,:) = max( x( j0:j1 ,:,:) );
		 case 4
			xr(i,:,:,:) = max( x( j0:j1 ,:,:,:) );
		 otherwise
			printf("[e] max_pooling_rows \n");
			exit -1;
	  endswitch
	  
   endfor
   
   retval=xr;

endfunction



% split_long_rows(0) ; % set nowrap
% x=rand([9,9,1,4]) ;
% %xr=max_pooling_rows(x,2) ;
% 
% w{1}{1}=1;
% w{1}{2}=0;
% 
% cfg{2}{2}=[2,2];
% 
% xr = layer_run_mpool( x, w, cfg, 2 );
% 
% size(x)
% size(xr)
 
 
