%!octave 
%
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: cn_trn_gd.m,v 1.13 2016/04/21 14:06:58 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% обучение ИНС методом градиентного спуска 
%
% L={Xl,Yl} - учебный набор
% С={Xc,Yc} - контрольный набор
% W - начальная матрица весов
% cfg - описание параметров сети сети
% trn_par - параметры обучения
%
function [W,er_hist]=cn_trn_gd(L,C,W,cfg,trn_par),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации

   min_err = trn_par{1} ; % порог допустимой ошибки обучения
   max_iter = trn_par{2} ; % максимальное число эпох
   er_win_size = trn_par{3} ;  % размер окна контроля за ошибкой
   trsp = trn_par{4} ; % параметр скорости обучения
   moment_par = trn_par{5} ; % параметр момента 
   reg_par = trn_par{6} ; % параметр регуляризации
   batch_par = trn_par{7} ; % доля учебных примеров в одном пакете от общего количества примеров
   print_step = trn_par{8} ;% печатаем значение ошибки каждые print_step эпох 

   X=L{1} ; y=L{2} ; % учебный набор
   num_examples=size(y,2) ; % количество объектов в учебном наборе
   batch_size=max(2,round(num_examples*batch_par)); % количество учебных примеров в одном пакете 

   tname="градиентный спуск";
   printf("[i] обучение методом %s\n",tname);
   save("-text","result/tname.txt", "tname" ) ;
 
   printf("[i] размер пакета (minibatch) %i / %i \n",batch_size,num_examples);

   er_hist = [er_win_size:-1:1]*100.0 ; % история значений ошибки обучения

   % случайный пакет из учебных примеров
   batch_idx = randperm(num_examples)(1:batch_size); 
   Lr = { X(:,:,:,batch_idx) , y(:,batch_idx) };

   g = cn_backprop_grad(Lr,W,cfg); % значение градиента
   g = cn_weight_norm(g,cfg);
   dW = weight_change(g,W,W,trsp,0.0,0.0,cfg); % изменение весов 
   W = cn_weight_add(W,dW,1.0,-1.0,cfg); % корректируем веса 

   for iter=1:max_iter,

	  % вычисляем ошибку сети на контрольном наборе
	  o = cn_run(C{1},W,cfg) ;
	  er_hist(end+1) = cn_loss(o,C{2},cfg) ;

	  if mod(iter,print_step)<1, % печатаем значение ошибки print_step эпох 
		 printf("      %i/%i : %f / %f   , %f     \r",iter,max_iter,er_hist(end),min_err,trsp ) ; 
      endif
 	 
	  if trn_stop_condition(er_hist,min_err,er_win_size), break ; endif % условия остановки обучения
 
	  % случайный пакет из учебных примеров
      batch_idx = randperm(num_examples)(1:batch_size); 
	  Lr = { X(:,:,:,batch_idx) , y(:,batch_idx) };
      
	  g = cn_backprop_grad(Lr,W,cfg); % значение градиента
	  g = cn_weight_norm(g,cfg);

	  dW = weight_change(g,W,dW,trsp,reg_par,moment_par,cfg); % изменение весов 
	  W = cn_weight_add(W,dW,1.0,-1.0,cfg); % корректируем веса 
	 
      %trsp=trn_speed(trsp,er_hist); % скорость обучения (изменения весов)
      %trsp = max(trsp*0.8,8e-2); % скорость обучения (изменения весов)
	  %trsp = abs(cos(iter)/iter ) + 0.07 ;

   endfor

   printf("[i]   %i/%i : %f / %f   , %f     \n",iter,max_iter,er_hist(end),min_err,trsp ) ; 

   er_hist(1:er_win_size)=[]; % удаляем начальные значения ошибки из истории 

   if iter>(max_iter-1) , printf("[w] достигнут порог количества циклов обучения         \n") ; endif

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% вычисляем изменение весов
% 
% g - градиент функции ошибки
% dW - пред.изменение весов
% W - весовая матрица
% trsp - скорость обучения
% rpar - параметр регуляризации
% mpar - параметр момента
%
function retval = weight_change(g,W,dW,trsp,rpar,mpar,cfg),
   %   веса   dW{1} := trsp * ( g + rpar * W ) + mpar * dW  
   % сдвиги   dW{2} := trsp * g  + mpar * dW 

   r = {} ;
   for i=1:length(g),
	  r{i} = weight_change_layer(g,W,dW,trsp,rpar,mpar,cfg,i) ;
   endfor

   retval = r;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_change_layer(g,W,dW,trsp,rpar,mpar,cfg,n),
   ltp=1; % номер ячейки с типом слоя 

   switch cfg{n+1}{ltp}

	  case "mlp"
	     r = weight_change_layer_mlp(g{n},W{n},dW{n},trsp,rpar,mpar) ;

	  case "conv"
	     r = weight_change_layer_conv(g{n},W{n},dW{n},trsp,rpar,mpar) ;

	  case "mpool"
		 %r={1,0};
	     %r = weight_change_layer_mpool(g{n},W{n},dW{n},trsp,rpar,mpar) ;
	     r = weight_change_layer_mlp(g{n},W{n},dW{n},trsp,rpar,mpar) ;

	  otherwise
		 printf("[E] weight_add_layer : неизвестный тип слоя %s \n",cfg{n+1}{ltp}); 
		 exit -1;

   endswitch

   retval = r;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_change_layer_mlp(g,W,dW,trsp,rpar,mpar,cfg,n),
   r={};
   r{1} = trsp.*(g{1} + rpar.*W{1}) + mpar.*dW{1} ;
   r{2} = trsp.*g{2} + mpar.*dW{2} ;

   retval = r;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_change_layer_conv(g,W,dW,trsp,rpar,mpar,cfg,n),
   r={};
   for i=1:length(g{1}),
	  r{1}{i} = trsp.*(g{1}{i} + rpar.*W{1}{i}) + mpar.*dW{1}{i} ;
   endfor
   r{2} = trsp.*g{2} + mpar.*dW{2} ;

   retval = r;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=trn_speed(s,er),
   k=1.01; % коэффициент допустимого роста ошибки
   ri=1.01; % коэффициент увеличения скорости
   rd=0.99; % коэффициент уменьшения скорости

   d = er(end) > k*er(end-1) ;

   if( d > 0 ),  % если существенный рост ошибки
	  s*=rd; % то уменьшаем скорость обучения
   else 
	  s=min(s*ri,1.0); % иначе увеличиваем скорость обучения
   endif

   retval=s ;
endfunction


