%!octave 
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: cn_trn_qprop.m,v 1.5 2016/04/11 20:57:00 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% обучение ИНС методом quickProp
%
% L={Xl,Yl} - учебный набор
% С={Xc,Yc} - контрольный набор
% W - начальная матрица весов
% cfg - описание параметров сети сети
% trn_par - параметры обучения
%
function [W,er_hist]=cn_trn_qprop(L,C,W,cfg,trn_par),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации

   min_err = trn_par{1} ; % порог допустимой ошибки обучения
   max_iter = trn_par{2} ; % максимальное число эпох
   er_win_size = trn_par{3} ;  % размер окна контроля за ошибкой
   trsp0 = trn_par{4} ; % параметр скорости обучения
   mpar0 = trn_par{5} ; % параметр момента 
   rpar = trn_par{6} ; % параметр регуляризации
   batch_par = trn_par{7} ; % доля учебных примеров в одном пакете от общего количества примеров
   print_step = trn_par{8} ;% печатаем значение ошибки каждые print_step эпох 

   X=L{1} ; y=L{2} ; % учебный набор
   num_examples=size(y,2) ; % количество объектов в учебном наборе
   batch_size=max(2,round(num_examples*batch_par)); % количество учебных примеров в одном пакете 



   tname="quickProp";
   printf("[i] обучение методом %s\n",tname);
   save("-text","result/tname.txt", "tname" ) ;

   printf("[i] размер пакета (minibatch) %i / %i \n",batch_size,num_examples);

   er_hist = [er_win_size:-1:1]*100.0 ; % история значений ошибки обучения

   % случайный пакет из учебных примеров
   batch_idx = randperm(num_examples)(1:batch_size); 
   Lr = { X(:,:,:,batch_idx) , y(:,batch_idx) };

   g = cn_backprop_grad(Lr,W,cfg); % значение градиента
   g = cn_weight_norm(g,cfg);
   dW = cn_weight_add(g,g,trsp0,0.0,cfg); % изменение весов 
   
   W0 = W;
   W = cn_weight_add(W,dW,1.0,-1.0,cfg); % корректируем веса 

   % trsp = cn_weight_set(trsp0,cfg) ; % начальные значения  скорости обучения
   % mpar = cn_weight_set(mpar0,cfg) ; % начальные значения параметра момента

 
   for iter=1:max_iter,
	  % вычисляем ошибку сети на контрольном наборе
	  o = cn_run(C{1},W,cfg) ;
	  er_hist(end+1) = cn_loss(o,C{2},cfg) ;

	  if mod(iter,print_step)<1, % печатаем значение ошибки print_step эпох 
		 printf("      %i/%i : %f / %f        \r",iter,max_iter,er_hist(end),min_err) ; 
      endif
 	  
	  if trn_stop_condition(er_hist,min_err,er_win_size), break ; endif % условия остановки обучения

	  g0 = g; % сохраняем старое значение градиента ф-ции потери

	  % случайный пакет из учебных примеров
	  batch_idx = randperm(num_examples)(1:batch_size); 
	  Lr = { X(:,:,:,batch_idx) , y(:,batch_idx) };

	  g = cn_backprop_grad(Lr,W,cfg); % новое значение градиента ф-ции потери
	  g = cn_weight_norm(g,cfg);

	  mpar = moment_param(g,W,g0,W0,rpar,cfg); % параметры момента для каждого веса

	  trsp = trn_speed_param(g,dW,W,rpar,trsp0,cfg); % скорость обучения для каждого веса

	  dW = weight_change(g,W,dW,trsp,rpar,mpar,cfg);% изменение весов 

	  W0 = W ; % сохраняем старые значения весов
	  W = cn_weight_add(W,dW,1.0,-1.0,cfg); % корректируем веса 
   endfor

   printf("[i]   %i/%i : %f / %f        \n",iter,max_iter,er_hist(end),min_err) ; 

   er_hist(1:er_win_size)=[]; % удаляем начальные значения ошибки из истории 

   if iter>(max_iter-1) , printf("[w] достигнут порог количества циклов обучения         \n") ; endif

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% вычисляем изменение весов
% 
% g - градиент функции ошибки
% dW - пред.изменение весов
% W - весовая матрица
% trsp - скорость обучения
% rpar - параметр регуляризации
% mpar - параметр момента
%
function retval = weight_change(g,W,dW,trsp,rpar,mpar,cfg),
   %   веса   dW{1} := trsp * ( g + rpar * W ) + mpar * dW  
   % сдвиги   dW{2} := trsp * g + mpar * dW 

   r = {} ;
   for i=1:length(g),
	  r{i} = weight_change_layer(g,W,dW,trsp,rpar,mpar,cfg,i) ;
   endfor

   retval = r;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_change_layer(g,W,dW,trsp,rpar,mpar,cfg,n),
   ltp=1; % номер ячейки с типом слоя 

   switch cfg{n+1}{ltp}

	  case "mlp"
	     r = weight_change_layer_mlp(g{n},W{n},dW{n},trsp{n},rpar,mpar{n}) ;

	  case "conv"
	     r = weight_change_layer_conv(g{n},W{n},dW{n},trsp{n},rpar,mpar{n}) ;

	  case "mpool"
		 %r={1,0};
	     %r = weight_change_layer_mpool(g{n},W{n},dW{n},trsp{n},rpar,mpar{n}) ;
	     r = weight_change_layer_mlp(g{n},W{n},dW{n},trsp{n},rpar,mpar{n}) ;

	  otherwise
		 printf("[E] weight_add_layer : неизвестный тип слоя %s \n",cfg{n+1}{ltp}); 
		 exit -1;

   endswitch

   retval = r;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_change_layer_mlp(g,W,dW,trsp,rpar,mpar),
   r={};
   r{1} = trsp{1} .*( g{1} + rpar.*W{1}) + mpar{1}.*dW{1} ;
   r{2} = trsp{2} .* g{2} + mpar{2}.*dW{2} ;

   retval = r;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_change_layer_conv(g,W,dW,trsp,rpar,mpar),
   r={};
   for i=1:length(g{1}),

	  r{1}{i} = trsp{1}{i} .* ( g{1}{i}  + rpar.*W{1}{i} ) + mpar{1}{i}.*dW{1}{i} ;
   endfor
   r{2} = trsp{2} .* g{2} + mpar{2}.*dW{2} ;

   retval = r;
endfunction






% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% скорость обучения для каждого веса
%
function retval= trn_speed_param(g,dW,W,rpar,trsp0,cfg),
   % S = -dW.*(g+rpar.*W)
   % trsp = trsp0 : ( dW = 0 ) or ( S > 0 )
   % trsp = 0     :  other

   S = cn_weight_add( g, W, 1.0, rpar, cfg ) ;
   S = cn_weight_mult( dW, S, -1.0, cfg ) ;

   % m = ( dW = 0 ) or ( S > 0 )
   m={};
   for i=1:length(S),
	  m{i}=trn_speed_param_layer_mask(dW,S,cfg,i);
   endfor

   r = cn_weight_set(trsp0,cfg);

   r = cn_weight_mult(r,m,1.0,cfg) ;

   retval = r;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = trn_speed_param_layer_mask(dW,S,cfg,n),
   ltp=1; % номер ячейки с типом слоя 

   switch cfg{n+1}{ltp}

	  case "mlp"
	     r = trn_speed_param_layer_mask_mlp( dW{n},S{n} );

	  case "conv"
	     r = trn_speed_param_layer_mask_conv( dW{n},S{n} );

	  case "mpool"
		 %r={0,0};
	     %r = trn_speed_param_layer_mask_mpool( dW{n},S{n} );
	     r = trn_speed_param_layer_mask_mlp( dW{n},S{n} );

	  otherwise
		 printf("[E] trn_speed_param_layer_mask : неизвестный тип слоя %s \n",cfg{n+1}{ltp}); 
		 exit -1;

   endswitch

   retval = r;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = trn_speed_param_layer_mask_mlp(dW,S),
   % m = ( dW = 0 ) or ( S > 0 )
   r={};
   r{1} = (dW{1} == 0.0) | (S{1} > 0);
   r{2} = (dW{2} == 0.0) | (S{2} > 0);

   retval = r;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = trn_speed_param_layer_mask_conv(dW,S),
   % m = ( dW = 0 ) or ( S > 0 )
   r={};
   for i=1:length(dW{1}),
	  r{1}{i} = (dW{1}{i} == 0.0) | (S{1}{i} > 0);
   endfor
   r{2} = (dW{2} == 0.0) | (S{2} > 0);

   retval = r;
endfunction







% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% параметры момента для каждого веса
%
function retval = moment_param(g,W,g0,W0,rpar,cfg),
   mpar_max=1.75;

   % S  = g+rpar*W
   % dS = S(t-1) - S(t)
   % b = S / dS
   % a = - S * dW * b

   % mpar = mpar_max : ( b > mpar_max ) or ( a < 0 )
   % mpar = b        :  other


   S  = cn_weight_add( g, W, 1.0, rpar, cfg ); % S=g+rpar*W;
   S0 = cn_weight_add(g0, W0,1.0, rpar, cfg ); 

   dS = cn_weight_add(S0, S, 1.0, -1.0, cfg );   % ds = s0 - s

   dW = cn_weight_add(W0, W, 1.0, -1.0, cfg );   % dW = W0 - W

   B = cn_weight_div(S,dS,1.0,cfg); % B = S./dS

   A = cn_weight_mult(S,dW,1.0,cfg); % (S.*dW.*B)
   A = cn_weight_mult(A,B,1.0,cfg); 

   r = {};
   for i=1:length(g),
	  r{i} = moment_param_layer(A,B,mpar_max,cfg,i);
   endfor

   retval = r;
endfunction




% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = moment_param_layer(A,B,mpar_max,cfg,n),
   ltp=1; % номер ячейки с типом слоя 

   switch cfg{n+1}{ltp}

	  case "mlp"
	     r = moment_param_layer_mlp(A{n},B{n},mpar_max);

	  case "conv"
	     r = moment_param_layer_conv(A{n},B{n},mpar_max);

	  case "mpool"
		 %r={0,0};
	     %r = trn_speed_param_layer_mpool( dW{n},S{n} );
	     r = moment_param_layer_mlp(A{n},B{n},mpar_max);

	  otherwise
		 printf("[E] weight_add_layer : неизвестный тип слоя %s \n",cfg{n+1}{ltp}); 
		 exit -1;

   endswitch

   retval = r;

endfunction



% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = moment_param_layer_mlp(A,B,mpar_max),
  % mpar = mpar_max : ( b > mpar_max ) or ( a < 0 )
  % mpar = b        :  other

   r={};
   
   m = (B{1} > mpar_max ) | (A{1} < 0);
   r{1} = m*mpar_max + not(m).*B{1} ;

   m = (B{2} > mpar_max ) | (A{2} < 0);
   r{2} = m*mpar_max + not(m).*B{2} ;

   retval = r;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = moment_param_layer_conv(A,B,mpar_max),
  % mpar = mpar_max : ( b > mpar_max ) or ( a < 0 )
  % mpar = b        :  other
   
   for i=1:length(B{1})
	  m = (B{1}{i} > mpar_max ) | (A{1}{i} < 0);
	  r{1}{i} = m.*mpar_max + not(m).*B{1}{i} ;
   endfor

   m = (B{2} > mpar_max ) | (A{2} < 0);
   r{2} = m.*mpar_max + not(m).*B{2} ;

   retval = r;
endfunction



