%!octave 
%
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: cn_trn_rprop.m,v 1.7 2016/04/11 20:57:00 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% обучение ИНС методом rProp (resilient back propagation)
%
% L={Xl,Yl} - учебный набор
% С={Xc,Yc} - контрольный набор
% W - начальная матрица весов
% cfg - описание параметров сети сети
% trn_par - параметры обучения
%
function [W,er_hist]=cn_trn_rprop(L,C,W,cfg,trn_par),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации

   min_err = trn_par{1} ; % порог допустимой ошибки обучения
   max_iter = trn_par{2} ; % максимальное число эпох
   er_win_size = trn_par{3} ;  % размер окна контроля за ошибкой
   trsp0 = trn_par{4} ; % параметр скорости обучения
   moment_par = trn_par{5} ; % параметр момента 
   reg_par = trn_par{6} ; % параметр регуляризации
   batch_par = trn_par{7} ; % доля учебных примеров в одном пакете от общего количества примеров
   print_step = trn_par{8} ; % печатаем значение ошибки print_step эпох

   %num_examples=size(L{2},2) ; % количество объектов в учебном наборе
   %batch_size=round(num_examples*batch_par); % количество учебных примеров в одном пакете 

   
   X=L{1} ; y=L{2} ; % учебный набор
   num_examples=size(y,2) ; % количество объектов в учебном наборе
   batch_size=max(2,round(num_examples*batch_par)); % количество учебных примеров в одном пакете 


   %er_win_size = 5; % размер окна контроля за ошибкой
   %trsp0 = 5e-3; % скорость обучения (изменения весов)
   %reg_par = 1e-4 ; % параметр регуляризации
   %moment_par = 5e-3; % параметр момента 

   tname="rProp";
   printf("[i] обучение методом %s\n",tname);
   save("-text","result/tname.txt", "tname" ) ;

   printf("[i] размер пакета (minibatch) %i / %i \n",batch_size,num_examples);

   er_hist = [er_win_size:-1:1]*100.0 ; % история значений ошибки обучения

 
   % случайный пакет из учебных примеров
   batch_idx = randperm(num_examples)(1:batch_size); 
   Lr = { X(:,:,:,batch_idx) , y(:,batch_idx) };

   g = cn_backprop_grad(Lr,W,cfg); % значение градиента
   g = cn_weight_norm(g,cfg);
   dW = cn_weight_add(g,W,trsp0,0.0,cfg); % изменение весов 
   W = cn_weight_add(W,dW,1.0,-1.0,cfg); % корректируем веса 
   trsp = cn_weight_set(trsp0,cfg) ; % начальные значения скорости обучения для каждого параметра

   for iter=1:max_iter,
	  o = cn_run(C{1},W,cfg); % вычисляем ошибку сети на контрольном наборе
	  er_hist(end+1) = cn_loss(o,C{2},cfg); % вычисляем ошибку сети на контрольном наборе

	  if mod(iter,print_step)<1, % печатаем значение ошибки print_step эпох 
		 printf( "      %i/%i : %f / %f     \r",iter,max_iter,er_hist(end),min_err ) ; 
      endif
	  if trn_stop_condition(er_hist,min_err,er_win_size), break ; endif % условия остановки обучения

	  g0 = g ; % сохраняем старое значение градиента 

	  % случайный пакет из учебных примеров
	  batch_idx = randperm(num_examples)(1:batch_size); 
	  Lr = { X(:,:,:,batch_idx) , y(:,batch_idx) };

	  g = cn_backprop_grad(Lr,W,cfg); % новое значение градиента
	  g = cn_weight_norm(g,cfg);

	  trsp = trsp_param(g,g0,trsp,cfg); % вычисляем параметр скорости обучения для каждого веса сети

	  dW = weight_change(g,W,dW,trsp,reg_par,moment_par,cfg);% изменение весов 
	  W = cn_weight_add(W,dW,1.0,-1.0,cfg); % корректируем веса 

   endfor

   printf("[i]   %i/%i : %f / %f     \n",iter,max_iter,er_hist(end),min_err) ; 

   er_hist(1:er_win_size)=[]; % удаляем начальные значения ошибки из истории 

   if iter>(max_iter-1) , printf("[w] достигнут порог количества циклов обучения         \n") ; endif

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval= trsp_param(g,g0,trsp,cfg),
   t_min = 1e-5 ;
   t_max = 20.0 ;
   a = 1.2 ;
   b = 0.5 ;
  
   S = cn_weight_mult(g,g0,1.0,cfg) ;
   
   R = {};
   for i=1:length(S),
	  R{i} = trsp_param_layer(S,trsp,a,b,t_max,t_min,cfg,i);
   endfor

   retval = R ;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = trsp_param_layer(S,trsp,a,b,t_max,t_min,cfg,n),
   ltp=1; % номер ячейки с типом слоя 

   switch cfg{n+1}{ltp}

	  case "mlp"
	     r = trsp_param_layer_mlp(S{n},trsp{n},a,b,t_max,t_min) ;

	  case "conv"
	     r = trsp_param_layer_conv(S{n},trsp{n},a,b,t_max,t_min) ;

	  case "mpool"
	     %r = trsp_param_layer_mpool(S{n},trsp{n},a,b,t_max,t_min) ;
		 r = { 0.0, 0.0 } ;

	  otherwise
		 printf("[E] trsp_param_layer: неизвестный тип слоя %s \n",cfg{n+1}{ltp}); 
		 exit -1;

   endswitch

   retval = r;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = trsp_param_layer_mlp(S,trsp,a,b,t_max,t_min),
   % S == 0
   r = trsp ; 

   % r := min(t_max, trsp * a ) : S > 0
   for j=1:2,
	  m = ( S{j} > 0.0);
	  t = min( a * trsp{j}, t_max ) ;
	  r{j} = trsp{j}.*not(m) + t .* m ; 
   endfor

   % R := max(t_min, trsp * b ) : S < 0
   for j=1:2,
	  m = ( S{j} < 0.0);
	  t = max( b * trsp{j}, t_min ) ;
	  r{j} = trsp{j}.*not(m) + t .* m ; 
   endfor

   retval = r;
  
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = trsp_param_layer_conv(S,trsp,a,b,t_max,t_min),
   % S == 0
   r = trsp ; 

   % для весов

   for i=1:length(S),
	  % r := min(t_max, trsp * a ) : S > 0
	  m = ( S{1}{i} > 0.0);
	  t = min( a * trsp{1}{i}, t_max ) ;
	  r{1}{i} = trsp{1}{i}.*not(m) + t .* m ; 

	  % R := max(t_min, trsp * b ) : S < 0
	  m = ( S{1}{i} < 0.0);
	  t = max( b * trsp{1}{i}, t_min ) ;
	  r{1}{i} = trsp{1}{i}.*not(m) + t .* m ; 

   endfor

   % для сдвигов

   % r := min(t_max, trsp * a ) : S > 0
   m = ( S{2} > 0.0);
   t = min( a * trsp{2}, t_max ) ;
   r{2} = trsp{2}.*not(m) + t .* m ; 

   % R := max(t_min, trsp * b ) : S < 0
   m = ( S{2} < 0.0);
   t = max( b * trsp{2}, t_min ) ;
   r{2} = trsp{2}.*not(m) + t .* m ; 

   retval = r;
  
endfunction




% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% вычисляем изменение весов
% 
% g - градиент функции ошибки
% dW - пред.изменение весов
% W - весовая матрица
% trsp - скорость обучения
% rpar - параметр регуляризации
% mpar - параметр момента
%
function retval = weight_change(g,W,dW,trsp,rpar,mpar,cfg),
   %   веса   dW{1} := trsp * ( sign(g) + rpar * W ) + mpar * dW  
   % сдвиги   dW{2} := trsp * sign(g)  + mpar * dW 

   r = {} ;
   for i=1:length(g),
	  r{i} = weight_change_layer(g,W,dW,trsp,rpar,mpar,cfg,i) ;
   endfor

   retval = r;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_change_layer(g,W,dW,trsp,rpar,mpar,cfg,n),
   ltp=1; % номер ячейки с типом слоя 

   switch cfg{n+1}{ltp}

	  case "mlp"
	     r = weight_change_layer_mlp(g{n},W{n},dW{n},trsp{n},rpar,mpar) ;

	  case "conv"
	     r = weight_change_layer_conv(g{n},W{n},dW{n},trsp{n},rpar,mpar) ;

	  case "mpool"
	     %r = weight_change_layer_mpool(g{n},W{n},dW{n},trsp{n},rpar,mpar) ;
		 %r={1,0};
	     r = weight_change_layer_mlp(g{n},W{n},dW{n},trsp{n},rpar,mpar) ;

	  otherwise
		 printf("[E] weight_add_layer : неизвестный тип слоя %s \n",cfg{n+1}{ltp}); 
		 exit -1;

   endswitch

   retval = r;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_change_layer_mlp(g,W,dW,trsp,rpar,mpar),
   r={};
   r{1} = trsp{1} .*( sign( g{1} ) + rpar.*W{1}) + mpar.*dW{1} ;
   r{2} = trsp{2} .* sign( g{2} ) + mpar.*dW{2} ;

   retval = r;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_change_layer_conv(g,W,dW,trsp,rpar,mpar),
   r={};
   for i=1:length(g{1}),
	  r{1}{i} = trsp{1}{i} .* ( sign( g{1}{i} ) + rpar.*W{1}{i}) + mpar.*dW{1}{i} ;
   endfor
   r{2} = trsp{2} .* sign( g{2} ) + mpar.*dW{2} ;

   retval = r;
endfunction


