%!octave 
%
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: cn_weight_div.m,v 1.3 2016/04/10 20:31:02 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% (w1 ./ w2) .* a
%
function retval = cn_weight_div(w1,w2,a,cfg),
   layers_numb = length(w1) ; % количество обрабатывающих слоёв 

   r={};
   for i=1:layers_numb,
      r{i} = weight_div_layer(w1,w2,a,cfg,i) ;
   endfor

   retval = r ;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_div_layer(w1,w2,a,cfg,n),
   ltp=1; % номер ячейки с типом слоя 

   switch cfg{n+1}{ltp}

	  case "mlp"
	     r = weight_div_mlp( w1{n},w2{n},a );

	  case "conv"
	     r = weight_div_conv( w1{n},w2{n},a );

	  case "mpool"
	     % r = { 1.0, 0.0 } ;
	     r = weight_div_mlp( w1{n},w2{n},a );

	  otherwise
		 printf("[E] weight_div_layer : неизвестный тип слоя %s \n",cfg{n+1}{ltp}); 
		 exit -1;

   endswitch

   retval = r;

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_div_mlp( w1n,w2n,a ),
   r={};

   d = w2n{1}+not(w2n{1})*1e-7 ;
   r{1} = ( w1n{1} ./ d ) .* a ;

   d = w2n{2}+not(w2n{2})*1e-7 ;
   r{2} = ( w1n{2} ./ d ) .* a ;

   retval = r ;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_div_conv( w1n,w2n,a ),
   r={};
   for i=1:length(w1n{1}),
	  d = w2n{1}{i}+not(w2n{1}{i})*1e-7 ;
	  r{1}{i} = ( w1n{1}{i} ./ d ) .* a ;
   endfor
     
   d = w2n{2}+not(w2n{2})*1e-7 ;
   r{2} = ( w1n{2} ./ d ) .* a ;

   retval = r ;
endfunction



