%!octave 
%
%      
%      
%       OS : GNU/Linux 4.3.3-3-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: cn_init.m,v 1.6 2016/04/08 13:40:30 mechanoid Exp mechanoid $

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% инициализация параметров сети
%
% cfg - конфигурация сети 
%
function retval = cn_weight_set( a, cfg ),
   lsz=2; % номер ячейки с размером слоя 
   
   s = cfg{1}{lsz}; % размер первого (входного) слоя сети

   W={}; % веса сети
   for i=2:length(cfg), % для всех слоёв
	  [ Wi, s ] = set_layer( a, cfg,i, s ) ;
	  W{end+1}=Wi;
   endfor

   retval = W ;

endfunction 


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% инициализация параметров слоя
%
% cfg - конфигурация сети 
% n - номер слоя
% W - параметры предыдущих слоёв
% s - размер предыдущего слоя
%
function [ W, s ]= set_layer( a, cfg, n, s ),
   ltp=1; % номер ячейки с типом слоя 

   switch cfg{n}{ltp}

   	  case "conv"
		 [ W,s ] = set_layer_conv(a, cfg, n, s );

	  case "mpool"
		 [ W,s ] = set_layer_mpool(a, cfg, n, s );

	  case "mlp"
	     [ W,s ] = set_layer_mlp(a, cfg, n, s );
	  
	  otherwise
		 printf("[E] неизвестный тип слоя %s \n",cfg{n}{ltp}); 
		 exit -1;

   endswitch


endfunction 

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function [ W, s ] = set_layer_mlp(a, cfg, n, s ),
   lsz=2; % номер ячейки с размером слоя 

   sz_pl = prod(s); % количество нейронов в пред.слое

   W={ ones( [ cfg{n}{lsz}, sz_pl ] ) .* a ,  ones(cfg{n}{lsz},1).* a  };

   s = cfg{n}{lsz}; % размер слоя

endfunction 

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function [ W, s ] = set_layer_mpool(a, cfg, n, s ),
   lsz=2; % номер ячейки с размером слоя 

   numb_maps = s(3);
 
   W = { ones( [ 1,1,numb_maps ] ).*a, ones( [ 1,1,numb_maps ]).*a   } ;
   
   % W = { 1.0, 0.0 } ; % этот слой выполняет только выборку максимальных элементов по ячейкам

   s(1:2) = ceil(s(1:2)./cfg{n}{lsz}) ;

endfunction 

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% инициализация параметров свёрточного слоя
%
% cfg - конфигурация сети 
% n - номер слоя
% W - параметры предыдущих слоёв
% s - размер предыдущего слоя
%
function [ W, s ] = set_layer_conv(a, cfg, n, s ),
   lsz = 2 ; % номер ячейки cfg с размером слоя 
   lmp = 4 ; % номер ячейки cfg с матрицей связей слоёв

   link_map = link_numb_seq( cfg{n}{lmp} ) ; % номера входных карт для частичных сумм

   map_out_numb = length(link_map); % количество результирующих карт
   sz_kern = cfg{n}{lsz} ; % размер ядра свёртки
   
   % одна карта формируется свёрткой частичной суммы входных карт и сдвигом: act( conv(psum(m),k_i)+b )

   k={}; % ядра свёрток
   for i=1:map_out_numb, % для всех выходных карт
      kern_numb = length(link_map{i} ) ; % количество ядер свёртки
	  k{i} = ones( [ sz_kern, kern_numb ] ) .* a ;
   endfor

   W = { k, ones( [ 1, 1, map_out_numb ] ) .* a } ;

   sz_map = s(1:2) - sz_kern + 1 ; % размер выходной карты

   s = [ sz_map, map_out_numb ] ; % размеры слоя n

endfunction 




