%!octave 
%
%      
%       OS : GNU/Linux 3.18.4-1-ARCH 
% COMPILER : GNU Octave, version 3.8.2
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: make_data_set.m,v 1.4 2016/02/08 17:22:24 mechanoid Exp mechanoid $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% разобрать полный набор данных {X,N} на учебный L={Xl,Yl} и контрольный C={Xc,Yc} наборы
%
% X - полный набор образов
% N - метки классов для набора X
% k - доля (0<k<1) примеров для использования от общего количества
% p - доля (0<p<1) учебных примеров  (остальные - контрольные примеры)
%
function [ L C ] = make_data_set(X,N,k,p),
   sz_img=size(X)(1:2); % размер образа
   %cl_num=length(unique(sort(N))); % количество классов
   cl_num=10;% количество классов

   printf("[i] всего примеров: %i   \n", length(N));

   if k<=0, % используем весь набор данных целиком
      printf("[i] используем полный набор данных \n");
      Y=[];
      for i=1:length(N),
	     nc=N(i); % номер класса для примера i
		 Y=[ Y ; zeros(1,nc) 1 zeros(1,(cl_num-1)-nc) ] ; % выход сети для примера i
	  endfor
	  L=C={X,Y};

	  return;
   endif


   % используем часть набора данных
   Xl=Yl=Xc=Yc=[];

   printf("[i] используем %.4f всего набора данных \n",k);

   for i=0:cl_num-1, % для каждого класса
      printf("[i] класс N %i   \r",i+1);

	  idx=find(N==i) ; % номера примеров для класса i

	  sz_idx=length(idx) ; % количество примеров класса i
	  idx=idx(randperm(sz_idx)) ; % случайный порядок примеров класса i

	  sz_data=round(sz_idx*k); % количество примеров для выборки
	  sz_data=max(2,sz_data); 

	  sz_l=round(sz_data*p); % ...из них учебные 
	  sz_l=max(1,sz_l); 
	  if(sz_l>0), 
		 idx_l=idx(1:sz_l); % номера учебных примеров класса i
		 Xl=cat( 4, Xl, X(:,:,1,idx_l) ); % учебные примеры класса i
		 Yl=[ Yl ; zeros(sz_l,i) ones(sz_l,1) zeros(sz_l,(cl_num-1)-i) ] ; % выход сети {0,1} (учебный) для класса i 
	  endif

	  sz_c=sz_data - sz_l;     % ...остальные - контрольные
	  sz_c=max(1,sz_c); 
	  if(sz_c>0), 
		 idx_c=idx(sz_l+1:sz_l+sz_c); % номера контрольных примеров класса i
		 Xc=cat( 4, Xc, X(:,:,1,idx_c) ); % контрольные примеры класса i
		 Yc=[ Yc ; zeros(sz_c,i) ones(sz_c,1) zeros(sz_c,(cl_num-1)-i) ] ; % выход сети {0,1} (контрольный) для класса i 
	  endif

   endfor

   L={Xl,Yl};
   C={Xc,Yc};

   printf("[i] примеров: %i - учебных: %i,   контрольных: %i      \n",size(Yl,1)+size(Yc,1),size(Yl,1),size(Yc,1));

endfunction
